# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/020_Encoder_Export.ipynb.

# %% auto 0
__all__ = ['dummy']

# %% ../nbs/020_Encoder_Export.ipynb 1
from fastkafka._components.encoder.avro import (
    AvroBase,
    avro_decoder,
    avro_encoder,
    avsc_to_pydantic,
)
from ._components.encoder.json import json_decoder, json_encoder
from ._components.meta import export

__all__ = [
    "AvroBase",
    "avro_decoder",
    "avro_encoder",
    "avsc_to_pydantic",
    "json_decoder",
    "json_encoder",
]

# %% ../nbs/020_Encoder_Export.ipynb 3
@export("_dummy")
def dummy() -> None:
    pass
